{-# OPTIONS --without-K --sized-types #-}
import Agda.Builtin.Bool
open import Agda.Builtin.Size
open import Data.Fin.Base
open import Data.List.Base
import Agda.Builtin.Equality
open import Agda.Builtin.Nat

module nested where

 data term : (n : Nat) -> Set where
   Var : {n : _} -> Fin n -> term n
   App : {n : _} -> term n -> List (term n) -> term n

 -- Does not pass the termination checker
 -- subst : {k l : Nat} -> (σ : Fin k -> term l) (t : term k) -> term l
 -- subst σ (Var x) = σ x
 -- subst σ (App t x) = App (subst σ t) (map (subst σ) x)

 -- This relies on size-change termination
 mutual
   subst_aux : {k l : Nat} -> (σ : Fin k -> term l) (t : term k) -> term l
   subst_aux σ (Var x) = σ x
   subst_aux σ (App t x) = App (subst_aux σ t) (map_subst_aux σ x)

   map_subst_aux : {k l : Nat} -> (σ : Fin k -> term l) (t : List (term k)) -> List (term l)
   map_subst_aux σ [] = []
   map_subst_aux σ (a ∷ as) = subst_aux σ a ∷ map_subst_aux σ as

 -- This relies on size-change termination too, just a local where variant
 subst_aux' : {k l : Nat} -> (σ : Fin k -> term l) (t : term k) -> term l
 subst_aux' σ (Var x) = σ x
 subst_aux' σ (App t x) = App (subst_aux' σ t) (map_subst_aux' σ x)
   where map_subst_aux' : {k l : Nat} -> (σ : Fin k -> term l) (t : List (term k)) -> List (term l)
         map_subst_aux' σ [] = []
         map_subst_aux' σ (a ∷ as) = subst_aux' σ a ∷ map_subst_aux' σ as
