Set Warnings "-notation-overridden".

Inductive eq {A : Type} (x : A) : A -> Prop := eq_refl : x = x where "x = y" := (eq x y) : type_scope.

(** Equality is an equivalence relation and its elimination principle
    [eq_rect_dep]
    is a dependent version of the Leibniz substitution principle,
    called the J rule in type theory jargon: *)
(* begin hide *)
Scheme eq_rect_dep := Induction for eq Sort Type.

Check eq_rect_dep : forall A x (P : forall y : A, x = y -> Type) (p : P x (eq_refl x)),
    forall (y : A) (e : x = y), P y e.

(* end hide *)
(**
[[
eq_rect_dep : forall {A} x (P : forall y : A, x = y -> Type) (p : P x (eq_refl x)) (y : A) (e : x = y), P y e
]]

    Informally, this principle states that to prove a goal [P y e] depending on
    a term [y] and a proof of equality [x = y], it suffices to show the
    case where [y] is substituted by [x] and the equality by [eq_refl x
    : x = x]. So, only the case where [y] and [x] are the same need be considered.

    An application [eq_rect_dep x P p y e] computes to its single arm
    [p : P x (eq_refl x)] when [e] is [eq_refl x]. The canonicity
    property of the theory ensures that if [p : t = u] in the empty
    context (i.e. when [p], [t] and [u] are closed terms), then [t] and
    [u] are convertible and [p] is [eq_refl]. In other words,
    propositional equality _reflects_ convertibility in the empty context.
    It is however a much larger relation under context: equality proofs
    can be built from induction principles, and assumptions of equality
    type might be false (e.g. [0 = 1]). It is good to bear in mind these intuitions
    when working with the (seemingly trivial) identity type.
 *)