(* begin hide *)
Set Warnings "-notation-overridden".
From Equations Require Import Equations. From EquationsReloaded Require Import Fin.
Set Equations Transparent.
Require Import Utf8 Lia List Omega.
From EquationsReloaded Require Import nestedrec.
Local Open Scope term.
Hint Extern 100 => progress (cbv zeta in *) : typeclass_instances.
From EquationsReloaded Require Import Sized.
(* end hide *)

(** ** Size-based nested well-founded recursion

  The [term_size] function uses the oveloaded [list_size] above. *)

Equations term_size {n} : Sized (term n) by struct :=
  term_size (Var v) := 1;
  term_size (Lam t) := S (size t);
  term_size (App t l) := S (size t + size l).
(* begin hide *)
Existing Instance term_size.
(* end hide *)

(** Finally subst_term can be defined using recursion on sizes. *)

Equations? subst {k l} (σ : fin k -> term l) (t : term k) : term l by wf (term_size t) lt := {
  (Var v) [σ] => σ v;
  (Lam t) [σ] => Lam (t [extend_var σ]);
  (App t l) [σ] => App (t [σ]) (map_size l (fun t Inl => t [σ])) }
where "t [ σ ]" := (subst σ t) : term.
Proof. all:unfold size in *; simpl; lia. Defined.
Hint Rewrite @map_size_spec : subst.

(** An eliminator talking only about [map] can also be derived if desired. *)

Lemma subst_term'_elim_all :
  ∀ (P : ∀ k l : nat, (fin k → term l) → term k → term l → Prop),
  (∀ k l σ (f : fin k), P k l σ (Var f) (σ f)) →
  (∀ k l σ (t : term (S k)), P (S k) (S l) (extend_var σ) t (t [extend_var σ]) →
    P k l σ (Lam t) (Lam (t [extend_var σ]))) →
  (∀ k l σ (t : term k) (ts : list (term k)), P k l σ t (t [σ]) →
    Forall2 (P k l σ) ts (map (subst σ) ts) →
    P k l σ (App t ts) (App (t [σ]) (map (subst σ) ts))) →
  ∀ k l σ t, P k l σ t (t [σ]).
Proof. intros P ????. apply (subst_elim _); intros *; auto.
       intros Ht Hl; generalize (map_size_transform _ _ _ Hl); simp subst.
Qed.
