namespace nested

  open nat

  inductive fin : nat -> Type
  | finz {n : nat} : fin (succ n)
  | fins {n : nat} : fin n -> fin (succ n)

  open nested.fin
  #check finz

  -- inductive term : nat -> Type
  -- fails with nested occurrence 'list.{0} (nested.term n)' contains variables that are not parameters
  -- | Var : Π {n : nat}, fin n -> term n
  -- | Abs : Π {n : nat}, term (nat.succ n) -> term n
  -- | App : Π {n : nat}, term n -> list (term n) -> term n

  mutual inductive term, term_list
  with term : nat -> Type
  | var {n : nat} : fin n -> term n
  | abs {n : nat} : term (succ n) -> term n
  | app {n : nat} : term n -> term_list n -> term n
  with term_list : nat -> Type
  | tNil {n} : term_list n
  | tCons {n} : term n -> term_list n -> term_list n
  open term term_list

  set_option eqn_compiler.lemmas false
  def lift_fin : Π {k} (l : nat), fin k -> fin (succ k)
  | k 0 f := fins f
  | k (succ n) finz := finz
  | k (succ n) (fins f) := fins (lift_fin n f)

  mutual def lift, lift_terms
  with lift : Π {k}, term k -> nat -> term (succ k)
  | k (var f) l := var (lift_fin l f)
  -- does not termination check, although lift is structurally recursive on term, term_list
  | k (abs f) l := abs (lift f (succ k))
  | k (app f ts) l := app (lift f l) (lift_terms ts l)
  with lift_terms : Π {k}, term_list k -> nat -> term_list (succ k)
  | k tNil l := tNil
  | k (tCons t ts) l := tCons (lift t l) (lift_terms ts l)

  def extend_var : Π {k l}, (fin k -> term l) -> (fin (succ k) -> term (succ l))
  | k l σ finz := var finz
  | k l σ (fins f) := lift (σ f) 0

  mutual def subst, subst_terms
  with subst : Π {k l : nat} (σ : fin k -> term l), term k -> term l
  | k l σ (var f) := σ f
  -- | k l σ (abs f) := abs (subst (extend_var σ) f)
  | k l σ (app f ts) := app (subst σ f) (subst_terms σ ts)
  with subst_terms : Π {k l : nat} (σ : fin k -> term l), term_list k -> term_list l
  | k l σ tNil := tNil
  | k l σ (tCons t ts) := tCons (subst σ t) (subst_terms σ ts)



end nested
