(** printing <= $\Leftarrow$ *)
(** printing < $<$ *)
(** printing <=mul $\le_{m}$ *)
(** printing =mul= $=_{m}$ *)
(** printing {{ $\{$ *)
(** printing }} $\}$ *)
(* begin hide *)
Set Asymmetric Patterns.
Set Implicit Arguments.
Unset Strict Implicit.

Require Import Arith Omega.
Require Import Equations.Equations.

Require Import Fpred.Tactics.
Require Import Fpred.Definitions.
Require Import Fpred.General.
Require Import Fpred.Metatheory.
Require Import Fpred.Reduction.

Require Import Relation_Operators Inclusion Relation_Definitions.
Require Import CoLoR.Util.Multiset.MultisetNat.

Arguments direct_subset_red {gtA gtA_eqA_compat M N} _.
Arguments red_eq_direct {gtA gtA_transitive gtA_eqA_compat M N}.
Arguments mord_ext_r {gtA gtA_transitive} gtA_irreflexive {gtA_eqA_compat} _ _ _.

(* Avoid name clash. *)

Definition empty_env := Fpred.Definitions.empty.
(* end hide *)

Derive Signature for le.

(*** Order for hereditary substitution ****)

(** To show that hereditary substitution is well-defined, we must
   provide an order of termination. In our case, we will have a
   lexicographic combination of a multiset ordering on kinds.  To
   formalize this, we reuse CoLoR's %\cite{BlanquiCoLoR}% library of
   multisets and definition of the multiset order. Those are multisets
   on ordered types, here natural numbers with the usual ordering, which
   is well-founded. *)

Notation "X <mul Y" := (MultisetLt gt X Y) (at level 70).

Definition wf_multiset_order : well_founded (MultisetLt gt).
Proof.
  refine (@mord_wf gt _ _).

  apply wf_incl with lt. unfold gt, transp.
  red. trivial.
  apply lt_wf.
Defined.

(** The [kinds_of] function computes the multiset of kinds appearing in a type,
  which reduces to the bounds of universal quantifications. *)

Equations kinds_of (t : typ) : Multiset :=
kinds_of (tvar _)    => empty; kinds_of (arrow T U) => union (kinds_of T) (kinds_of U);
kinds_of (all k T)   => union {{k}} (kinds_of T).

(* begin hide *)
Lemma gt_trans : transitive gt.
Proof. refine gt_trans. Qed.
Hint Resolve gt_trans.

Lemma compat : forall x x' y y' : A, x =A= x' -> y =A= y' -> x > y -> x' > y'.
Proof. intros x x' y y' -> ->. trivial. Qed.
Hint Resolve compat.

Import Nat.

Lemma singleton_mult_eq : forall a a', {{a}} =mul= {{a'}} -> a = a'.
Proof.
  intros. repeat red in H. simpl in H. specialize (H a).
  destruct eqA_dec; try contradiction. simpl in H.
  destruct eqA_dec; try contradiction. simpl in H. destruct e0. reflexivity.
  discriminate. destruct eqA_dec; try discriminate. 
  elimtype False; apply n. reflexivity.
Qed.

Lemma singleton_mult_union_eq : forall a a' M,
    {{a}} =mul= {{a'}} + M -> (a = a' /\ M =mul= empty).
Proof.
  intros.
  assert (M =mul= empty) by (apply singleton_mult_union_empty in H; trivial).
  rewrite H0 in H. rewrite union_empty in H. apply singleton_mult_eq in H. intuition.
Qed.

Lemma union_empty_l M N : M + N =mul= empty -> M =mul= empty.
Proof.
  destruct M. simpl. intros. unfold empty. constructor.
  simpl. intros. red in H. red in H. simpl in H.
  red in H. simpl in H. specialize (H a). destruct eqA_dec. discriminate.
  elimtype False. apply n. reflexivity.
Qed.

Lemma union_empty2 M N : M + N =mul= empty -> M =mul= empty /\ N =mul= empty.
Proof.
  intros. assert (foo:=union_empty_l H). rewrite union_comm in H.
  pose (union_empty_l H). intuition.
Qed.

Lemma singleton_mult : forall a a',
                         {{a}} <mul {{a'}} -> a < a'.
Proof.
  intros. 
  do 2 red in H. rewrite red_eq_direct in H; eauto.
  destruct H. 
  apply notempty_member in H. destruct H as [N' [a0 H]].
  rewrite H in H0. rewrite union_comm in H0. rewrite <- union_assoc in H0.
  eapply singleton_mult_union_eq in H0. destruct H0. subst a'.
  apply union_empty2 in H3. destruct H3. rewrite H0 in H. rewrite H3 in H1.
  rewrite union_empty in H. rewrite union_comm in H1. rewrite union_empty in H1.
  specialize (H2 a). destruct H2. rewrite <- H1. auto with multisets.
  rewrite H in H2. apply member_singleton in H2. rewrite H2 in H4. auto.
Qed.

Require Import Morphisms.

Hint Unfold eqA : typeclass_instances.

Lemma mul_singleton M k : M <mul {{k}} <-> forall x, x in M -> k > x.
Proof.
  split; intros.
  do 2 red in H. rewrite red_eq_direct in H; eauto.


  destruct H. apply notempty_member in H. destruct H as [N' [a0 H]].
  rewrite H in H1. rewrite union_comm in H1. rewrite <- union_assoc in H1.
  eapply singleton_mult_union_eq in H1. destruct H1. subst a0.
  apply union_empty2 in H4. destruct H4. rewrite H4 in H2.
  rewrite union_comm, union_empty in H2.
  specialize (H3 x).
  destruct H3. rewrite <- H2. auto. 
  rewrite H in H3.
  rewrite H1 in H3. rewrite union_empty in H3.
  apply member_singleton in H3. rewrite H3 in H5. trivial.

  econstructor. econstructor; eauto.
  rewrite union_comm, union_empty. reflexivity.
  rewrite union_comm, union_empty. reflexivity.
Qed.

Lemma le_singleton_mul k M k' : M <mul {{k'}} -> {{k}} + M <mul {{max (k+1) k'}}.
Proof.
  intros.
  rewrite mul_singleton in H.
  rewrite mul_singleton.
  intros.
  apply member_union in H0. destruct H0.
  apply member_singleton in H0. rewrite H0.

  red. rewrite Nat.max_lt_iff. left; omega.
  red. rewrite Nat.max_lt_iff. right. apply H. auto.
Qed.

Lemma mul_union_l M N P : M <mul P -> M <mul P + N.
Proof.
  intros. do 2 red in H. do 2 red.
  apply direct_subset_red; eauto.
  rewrite red_eq_direct in H; eauto.

  destruct H.
  apply MSetGT with (X + N) Y Z; auto with multisets.
  rewrite union_assoc. rewrite H0. reflexivity.

  intros.
  destruct (H2 y H3). exists x; auto with multisets.
Qed.

Lemma empty_singleton k : empty <mul {{k}}.
Proof.
  rewrite mul_singleton. intros. depelim H.
Qed.

Equations mul_sum (n : nat) (M : Multiset) : Multiset :=
mul_sum  O    _ => empty;
mul_sum (S n) M => M + mul_sum n M.

Lemma mul_sum_sum n k M : mul_sum (n + k)%nat M =mul= mul_sum n M + mul_sum k M.
Proof.
  funelim (mul_sum n M); simpl; simp mul_sum. solve_meq. rewrite H; solve_meq.
Qed.

Require Import Morphisms Setoid.

Instance: Proper (meq ==> meq ==> iff) (MultisetLt gt).
Proof. reduce. apply MultisetGt_morph_equiv; auto. Qed.

Lemma gt_irrefl : RelUtil.irreflexive gt.
Proof. red. intros. omega. Qed.
Hint Resolve gt_irrefl.

(* end hide *)

(** Clearly, the singleton multiset built from any valid kind for [T]
  bounds the bag of kinds appearing in [T], according to the kinding
  rules. This is proved by induction on the kinding derivation: *)

Lemma kinds_of_kinded e T k : kinding e T k -> kinds_of T <mul {{ k }}.
Proof.
  induction 1. simpl. apply direct_subset_red; eauto.
  apply (@MSetGT _ _ _ {{k'}} empty empty); auto with multisets.
  intros. depelim H2.

  rewrite mul_singleton in IHkinding1, IHkinding2.
  rewrite mul_singleton. intros.
  simp kinds_of in H1. apply member_union in H1. destruct H1.
  specialize (IHkinding1 _ H1). red. rewrite Nat.max_lt_iff. left; omega.
  specialize (IHkinding2 _ H1). red. rewrite Nat.max_lt_iff. right; omega.

  rewrite mul_singleton in IHkinding.
  rewrite mul_singleton.
  intros. simp kinds_of in H0.
  apply member_union in H0.
  destruct H0.
  apply member_singleton in H0. rewrite H0.
  red; rewrite Nat.max_lt_iff. left; omega.
  red; rewrite Nat.max_lt_iff. right; eauto. apply IHkinding. auto.
Defined.

(** Kinds in a type are invariant by shifting or lifting. This is a simple example
  of a proof by functional elimination. The [T] argument and result of [kinds_of T] 
  get refined and we just need to simplify the right hand sides according to the
  definitions of [kinds_of] and [tshift], using rewriting not computation, and 
  finish by rewriting with the induction hypotheses.
*)

Lemma kinds_of_tshift X T : kinds_of (tshift X T) = kinds_of T.
(* begin show *)
Proof.
  funelim (kinds_of T); simp kinds_of tshift; auto; now rewrite H, ?H0.
Qed.
(* end show *)

(** For type substitution of [T] in [U] however, an exact arithmetic relation holds.
  We know that the multiset of kinds of the substituted type can appear 
  a finite number of times in the resulting type, along with the original 
  kinds of [U]. *)

Lemma kinds_of_tsubst e e' X T U k : env_subst X T e e' -> kinding e U k ->
  exists n : nat, kinds_of (tsubst U X T) =mul= kinds_of U + mul_sum n (kinds_of T).
Proof.
  intros. funelim (kinds_of U); simp tsubst kinds_of.

  - destruct lt_eq_lt_dec as [[L|R]|R]; simpl.
    * exists 0; reflexivity.
    * exists 1. simp mul_sum. rewrite union_empty. reflexivity.
    * exists 0; reflexivity.

  - depelim H2.
    do 2 match goal with
    | H : forall _, _,
      H1 : env_subst _ _ _ _,
      H2 : kinding _ _ _ |- _ => destruct (H _ _ _ _ _ H1 H2); clear H2
    end.
    exists (x + x0)%nat.
    rewrite H2, H3. rewrite mul_sum_sum. solve_meq.

  - depelim H1.
    destruct (H _ _ (S X) (tshift 0 T) k' (es_kind _ H0) H1).
    exists x. rewrite H2. rewrite kinds_of_tshift. reflexivity.
Qed.

(** This allows us to derive a general result about kindings of
   universal types: any well-kinded instance substitution produces a
   type with a strictly smaller bag of kinds. This is the central 
   result needed to show termination. In Stump's work, the measure
   considered was solely the depth of types, and only through the stricter
   kinding invariant could the order be shown well-founded. *)

Lemma kinds_of_tsubst_all e U k k' T : kinding e (all k U) k' ->
  kinding e T k -> kinds_of (tsubst U 0 T) <mul kinds_of (all k U).
Proof.
  intros.
  depelim H.
  destruct (kinds_of_tsubst (es_here H0) H).
  rewrite H1.
  simp kinds_of.
  rewrite union_comm.
  red. unfold transp.
  apply mord_ext_r; eauto. apply _.

  change (MultisetLt gt (mul_sum x (kinds_of T)) {{k}}).
  generalize (kinds_of_kinded H0).
  clear. intros; induction x. simpl.
  apply empty_singleton.

  simpl. rewrite (mul_singleton _ k).
  rewrite mul_singleton in IHx.
  intros. simp mul_sum in H0.
  apply member_union in H0. destruct H0.
  rewrite mul_singleton in H. eauto.
  eauto.
Qed.

(* begin hide *)

Require Import Wellfounded Relation_Definitions.
Require Import Relation_Operators Lexicographic_Product Wf_nat.
Arguments lexprod [A B].

Section MyLexicographic_Product.

  Variable A : Type.
  Variable B : Type.
  Variable leA : A -> A -> Prop.
  Variable eqA : A -> A -> Prop.
  Variable leB : B -> B -> Prop.

  Inductive lexprod : A * B -> A * B -> Prop :=
    | left_lex :
      forall (x x':A) (y:B) (y':B),
        leA x x' -> lexprod (x, y) (x', y')
    | right_lex :
      forall (x x':A) (y y':B),
        eqA x x' -> leB y y' -> lexprod (x, y) (x', y').

End MyLexicographic_Product.

Derive Signature for lexprod.

Require Import RelationClasses.
Section MyWfLexicographic_Product.
  Variable A : Type.
  Variable B : Type.
  Variable leA : A -> A -> Prop.
  Variable eqA : A -> A -> Prop.
  Variable leB : B -> B -> Prop.

  Notation LexProd := (lexprod leA eqA leB).
  Variable leA_equiv : Equivalence eqA.
  Variable leA_eqA : forall x x' y, eqA x x' -> leA y x' -> leA y x.

  Lemma acc_eqA_leA : forall (p : A * B) (x' : A), eqA (fst p) x' -> 
                                                 Acc LexProd p -> Acc LexProd (x', snd p).
  Proof.
    intros. revert x' H.
    induction H0. constructor. intros.
    apply H. simple inversion H2.
    subst y. injection H5; intros; subst. destruct x. constructor. simpl in *.
    apply leA_eqA with x'; auto.

    subst y. injection H6; intros; subst. clear H6.
    destruct x. simpl in *. constructor 2. transitivity x'; auto. symmetry; auto.
    auto.
  Qed.

  Lemma acc_A_B_lexprod :
    forall x:A, Acc leA x -> (well_founded leB) ->
                forall y:B, Acc leB y -> Acc LexProd (x, y).
  Proof.
    induction 1 as [x _ IHAcc]; intros H2 y.
    induction 1 as [x0 H IHAcc0]; intros.
    apply Acc_intro.
    destruct y as [x2 y1]; intro H6.
    simple inversion H6; intro.
    injection H1. injection H3. intros. subst. clear H1 H3.
    apply IHAcc; auto with sets.

    intros.
    injection H3. injection H4. intros. subst. clear H4 H3.
    pose (IHAcc0 _ H1).
    apply (@acc_eqA_leA (x, y1) x2). simpl. now symmetry. apply a.
  Qed.

  Theorem wf_lexprod :
    well_founded leA ->
    well_founded leB -> well_founded LexProd.
  Proof.
    intros wfA wfB; unfold well_founded.
    destruct a.
    apply acc_A_B_lexprod; auto with sets; intros.
  Defined.

End MyWfLexicographic_Product.
(* end hide *)

(** ** Definition of the measure. *)

(** We first define the depth of a type as being the number of universal
    quantifications and type variables in that type. *)

Equations depth (t : typ) : nat :=
depth (tvar  _)   => 1; depth (arrow T U) => (depth T + depth U)%nat;
depth (all   k U) => S (depth U).

(** Of course, it cannot be zero, which is useful since it allows
    to have [depth T < depth (arrow T U)], which will be needed to
    prove the well-foundedness of the hereditary substitution. *)

Lemma depth_nz t : 0 < depth t.
Proof.
  funelim (depth t); omega.
Qed.

(* begin hide *)
Lemma depth_tshift T : forall k, depth (tshift k T) = depth T.
Proof. intros; simpl; funelim (depth T); simp tshift depth; auto. Qed.

Definition MR {A B} (f : A -> B) (R : relation B) : relation A :=
  fun x y => R (f x) (f y).
(* end hide *)

(** The order that we will use on types is a lexicographical order on
    the multiset of kinds and the depth. As the first part of the
    lexicographic product is a multiset, and as those should not be
    compared with the Leibniz equality but rather a specific setoid
    equality, we defined a generalized notion of lexicographic product
    up-to an equivalence relation on the first component, here [meq]
    which represents multiset equality. *)

Definition relmd : relation (Multiset * nat) := lexprod (MultisetLt gt) meq lt.

(** It is well-founded, relying ultimately on the well-foundedness of [lt]. *)

Definition wf_relmd : well_founded relmd.
Proof.
  apply (wf_lexprod); unfold MR. 
  constructor; red; intros. reflexivity. now symmetry. etransitivity; eauto.

  intros. rewrite <- H in H0. auto.

  apply wf_multiset_order.
  apply lt_wf.
Defined.

(** The actual order is [relmd] _on_ the [kinds_of] and [depth] measures on types. *)

Definition order (x y : typ) := relmd (kinds_of x, depth x) (kinds_of y, depth y).

Definition wf_order : well_founded order.
Proof.  unfold order.
  apply wf_inverse_image. apply wf_relmd.
Defined.

(** It is well-founded and clearly transitive. *)
Lemma order_trans t u v : order t u -> order u v -> order t v.
Proof.
  intros H H'; unfold order in *; eauto. simpl in *.
  red in H. red in H'. red. depelim H; depelim H'.
  - constructor. red in H, H0; red; unfold transp in *; eauto using mord_trans.
  - left. rewrite <- H0; trivial.
  - left. rewrite H; trivial.
  - right. rewrite H; trivial. transitivity (depth u); trivial.
Qed.

(** As we expected, we can compare a type with an arrow on that type, on the left
    and on the right. *)

Lemma order_arrow_l : forall A B, order A (arrow A B).
Proof.
  intros A B; induction B.
  - constructor 2; red; simp kinds_of depth.
    + rewrite union_empty. reflexivity.
    + omega.
  - depelim IHB1.
    + constructor; red in H; red; simp kinds_of in *.
      rewrite union_assoc. apply mul_union_l. trivial.
    + depelim IHB2.
      * constructor; red in H; red; simp kinds_of in *.
        rewrite union_assoc. rewrite union_perm. apply mul_union_l. trivial.
      * constructor 2; red in H, H0, H1, H2; red; simp kinds_of depth in *.
        -- rewrite union_assoc. rewrite <- H. trivial.
        -- omega.
  - constructor; red; simp kinds_of. rewrite union_assoc.
    apply mul_union_l. constructor. econstructor.
    reflexivity. rewrite union_empty. reflexivity.
    intros. depelim H.
Qed.

Lemma order_arrow_r : forall A B, order B (arrow A B).
Proof.
  intros A B; induction A.
    - constructor 2; red; simp kinds_of depth; reflexivity.
    - depelim IHA1.
      + constructor; red in H; red; simp kinds_of in *.
        rewrite union_perm. apply mul_union_l. trivial.
      + depelim IHA2.
        * constructor; red in H; red; simp kinds_of in *.
          rewrite <- union_assoc. rewrite union_comm.
          apply mul_union_l. trivial.
        * constructor 2; red in H, H0, H1, H2; red; simp kinds_of depth in *.
          -- rewrite union_perm. rewrite <- H. rewrite union_comm. trivial.
          -- omega.
   - constructor; red; simp kinds_of. rewrite union_comm.
     rewrite union_assoc. apply mul_union_l. constructor. econstructor.
     reflexivity. rewrite union_empty. reflexivity.
     intros. depelim H.
Qed.

(** We also define the reflexive closure of this order. It will be useful
    to express the postcondition of the hereditary substitution function,
    as we will explain below. *)

Definition ordtyp := clos_refl order.
Derive Signature for clos_refl.
(* begin hide *)
Hint Unfold ordtyp.
Arguments wf_inverse_image [A B].

Lemma order_ordtyp_trans t u v : order t u -> ordtyp u v -> order t v.
Proof.
  unfold ordtyp.
  intros H H'. destruct H' as [|]. eapply order_trans; eauto. auto.
Qed.
(* end hide *)

(** Finally, we define the size of a term as usual. *)

Equations term_size (t : term) : nat :=
term_size (var  _)   => 0; term_size (abs  T t) => S (term_size t);
term_size (app  t u) => S (term_size t + term_size u);
term_size (tabs k t) => S (term_size t); term_size (tapp t U) => S (term_size t).

Definition wf_term_size : well_founded (MR term_size lt) := wf_inverse_image lt term_size lt_wf.

(** The hereditary substitution order is a lexicographic combination 
  of the order on the multisets of kinds in the substituted term's type,
  the number of universal quantifiers and type variables in the substituted
  term's type, and the term size of the substituend.
  In other words, with [U] the type of the substituted term and [t] the
  substituend, we first compare the multiset of kinds in [U], then the
  depth of [U], and ultimately the size of [t]. *)

Definition her_order : relation (typ * term) := 
  lexprod order (fun x y => kinds_of x = kinds_of y /\ depth x = depth y) (MR term_size lt).

Instance WF_her_order : WellFounded her_order.
Proof.
  apply wf_lexprod. 
  * constructor; red; intros.
  - auto. 
  - destruct H. split;symmetry; auto. 
  - destruct H, H0. split;etransitivity;eauto.

  * intros x x' y [H H'] H0; simpl in *. do 2 red in H0. do 2 red. simpl in *.
    depelim H0. constructor; auto. red in H0. now rewrite <- H in H0.
    red in H0, H1. right. rewrite <- H in H0. apply H0. rewrite <- H' in H1. 
    auto.
  * apply wf_order.
  * apply wf_term_size.
Defined.

(* begin hide *)
Hint Resolve wf_typ_weakening_var.

Hint Constructors typing.

Hint Resolve kinding_weakening_var kinding_strengthening_var wf_typ_remove_var.
Hint Resolve typing_weakening_var.
(* end hide *)

(** ** The model

  We now turn to the interpretation proper. We characterize the normal forms as a
 subset of the terms using mutually-inductive [normal] and [neutral] judgments.
 The plan is to show that the hereditary substitution function, when given two terms in 
 [normal] form will produce terms in [normal] form. We can already expect some complications
 as normal terms also include [neutral] ones... *)

Inductive normal : term -> Prop :=
| normal_abs T t : normal t -> normal (abs T t)
| normal_tabs k t : normal t -> normal (tabs k t)
| normal_neutral r : neutral r -> normal r
with neutral : term -> Prop :=
| neutral_var i : neutral (var i)
| neutral_app t n : neutral t -> normal n -> neutral (app t n)
| neutral_tapp t T : neutral t -> neutral (tapp t T).
(* begin hide *)
Derive Signature for normal.
Derive Signature for neutral.
(* end hide *)

(* begin hide *)
Hint Constructors normal neutral.
Hint Constructors env_subst.
Hint Resolve env_subst_wf_env.
(* end hide *)


(** A term [t] is said to be a canonical inhabitant of a type [T] in
    environment [e] if [e |- t : T] and [t] is in normal form. Our goal
    will be to show that every typeable term can be normalized to a 
    canonical one. *)

Definition canonical e t T := typing e t T /\ normal t.

(* begin hide *)
Definition interp_typing e t T : canonical e t T -> typing e t T := @proj1 _ _.
Definition interp_normal e t T : canonical e t T -> normal t := @proj2 _ _.

Hint Unfold canonical.
Hint Resolve interp_typing interp_normal.
(* end hide *)

(** We define a relation expressing that [n] is the interpretation of some arbitrary term [t]
  of type [T] and in environment [e]. *)
Definition interp e t T n := reds t n /\ canonical e n T.

(* begin hide *)
Lemma tsubst_preserves_normality : (forall n k T, normal n -> normal (subst_typ n k T))
with tsubst_preserves_neutrality : (forall n k T, neutral n -> neutral (subst_typ n k T)).
Proof.
  - destruct 1.
    * simp subst_typ. constructor. auto.
    * simp subst_typ. constructor. apply tsubst_preserves_normality. apply H.
    * apply normal_neutral. apply tsubst_preserves_neutrality. apply H.

  - destruct 1. 
    * simp subst_typ. constructor.
    * simp subst_typ. constructor. apply tsubst_preserves_neutrality.
      apply H.
      apply tsubst_preserves_normality. apply H0.
    * simp subst_typ. constructor. apply tsubst_preserves_neutrality. apply H.
Qed.

Lemma normal_shift n : normal n -> forall k, normal (shift k n)
with neutral_shift n : neutral n -> forall k, neutral (shift k n).
Proof.
  intros H; destruct H; intros; simp shift; constructor; eauto.
  intros H; destruct H; intros; simp shift; constructor; eauto.
Qed.

Lemma normal_shift_typ n : normal n -> forall k, normal (shift_typ k n)
with neutral_shift_typ n : neutral n -> forall k, neutral (shift_typ k n).
Proof.
  intros H; destruct H; intros; simp shift_typ; constructor; eauto.
  intros H; destruct H; intros; simp shift_typ; constructor; eauto.
Qed.

Hint Resolve tsubst_preserves_neutrality tsubst_preserves_normality.
Hint Resolve normal_shift neutral_shift.
Hint Resolve normal_shift_typ neutral_shift_typ.

Equations is_absterm (t : term) : Prop :=
is_absterm (abs  _ _) => True;
is_absterm  _         => False.
Lemma is_absterm_dec (t : term) : {is_absterm t} + {~ is_absterm t}.
Proof. destruct t; simp is_absterm; auto. Qed.

Equations is_tabsterm (t : term) : Prop :=
is_tabsterm (tabs  _ _) => True;
is_tabsterm  _          => False.
Lemma is_tabsterm_dec (t : term) : {is_tabsterm t} + {~ is_tabsterm t}.
Proof. destruct t; simp is_tabsterm; auto. Qed.

Notation pair := Datatypes.pair.
Unset Implicit Arguments.
(* end hide *)

(** ** Hereditary substitution

  As we said, hereditary substitution takes two terms [t] and [u] in [normal]
  form and returns a term which is the result of substituting [u] in [t] at
  some index. From a purely algorithmic point of view, we only need [t], [u]
  and the index [X] to compute the result of this function. However, we need
  more to prove its correctness.
  
  First of all, the well-founded order that we use to justify its termination
  is an order on the type of the substituted and on the substituend, which
  is why the function [hsubst] also takes as an argument the type
  of the substituted term.
  
  We then need a typing environment for [t] and [u], which is not useful from
  a computational point of view but will serve to prove the termination and
  the correctness of [hsubst]. To this effect, we will decorate
  the function with a precondition and a postcondition. We define those
  in the universe of propositions to underline the fact that they are not
  useful in a computational way. *)

Definition pre (T : typ) (t : term) (u : term) (X : nat) (p : env * typ) : Prop :=
    (get_var (fst p) X = Some T /\ canonical (fst p) t (snd p)
     /\ canonical (remove_var (fst p) X) u T).

(* Definition pre (t : typ * term) (u : term) (X : nat) (p : env * typ) : Prop := *)
(*   let (T, t) := t in True. *)
(*   let (e, T') := p in *)
(*     (get_var e X = Some T /\ canonical e t T' /\ canonical (remove_var e X) u T). *)

(** There is one subtle point in the formulation of the
    postcondition. When we substitute in an application [app t1 t2], it
    may be that the result of substituting in [t1] is an abstraction
    [abs T t]. If that's the case, to preserve the invariant that the
    result of [hsubst] is in normal form, we have to call
    again [hsubst] to perform the beta-reduction. However to
    do this, we need to know that the type of [t1] is smaller than the
    type of the substituted term.  Note that we need to add a
    side-condition to this property, which is not always true: if the
    substituted variable did not appear at all, then there is no reason
    to have any relation between those types. There is a relation only if
    the original term was not an abstraction but the substituted
    term is. *)

Equations is_abs (t : term) : Prop :=
is_abs (abs  _ _) => True; is_abs (tabs _ _) => True; is_abs  _         => False.

(* begin hide *)
Lemma is_abs_dec (t : term) : {is_abs t} + {~ is_abs t}.
Proof. destruct t; simp is_abs; auto. Qed.

Lemma noabs_neutral (t : term) : normal t -> ~ is_abs t -> neutral t.
Proof. induction 1; intros; simp is_abs in *; tauto. Qed.

Lemma neutral_noabs (t : term) : neutral t -> ~ is_abs t.
Proof. induction 1; simp is_abs in *; tauto. Qed.
(* end hide *)

Definition post (U : typ) (t : term) (u : term) (X : nat) (r : term) (p : env * typ) : Prop :=
  interp (remove_var (fst p) X) (subst t X u) (snd p) r /\
  (~ is_abs t -> is_abs r -> ordtyp (snd p) U).

(** The Program mode proposed by %\Coq% interacts nicely with %\Equations%, in
    that it allows us to just return a term, and provide later the
    postcondition, thanks to subtyping of subset types. In the same way,
    we can treat a value returned by a call as if it was just the
    term. We use a standard encoding for the ghost [p : env * typ] variable.
    The (noind) option disables the generation of the graph and elimination principle
    for the function, its type and computational behavior is all we need here. *)

Set Program Mode.

Arguments exist [A] [P] x p.

Equations hsubst (U : typ) (t: term) (u : term) (X : nat)
         (P : exists (p : env * typ), pre U t u X p) :
  {r : term | forall (p : env * typ), pre U t u X p -> post U t u X r p} by wf (U, t) her_order :=
hsubst U (var i) u X P with lt_eq_lt_dec i X => {
    | inleft (right p) => u; | inleft (left p) => var i;
    | inright p => var (pred i) };
hsubst U (abs T t) u X P => abs T (hsubst U t (shift 0 u) (S X) _) ;
hsubst U (tabs k t) u X P => tabs k (hsubst (tshift 0 U) t (shift_typ 0 u) X _) ;
hsubst U (tapp t T) u X P with hsubst U t u X _ => {
  | exist (tabs k t') Habs => subst_typ t' 0 T;
  | exist r Hr => tapp r T };
hsubst U (app t1 t2) u X P with hsubst U t2 u X _ => {
  | exist r2 P2 with hsubst U t1 u X _ => {
    | exist (abs T' t') P1 => hsubst T' t' r2 0 _;
    | exist r1 P1 => app r1 r2 } }.

(* 6 recursive calls --> 12 obligations *)
(* 15 right-hand-sides --> 15 obligations *)

(* begin hide *)

Ltac destruct_canonical :=
  match goal with
  | H : interp _ _ _ _ |- _ => destruct H
  | H : canonical _ _ _ |- _ => destruct H
  | H : _ /\ _ |- _ => destruct H
  | H : exists _, _ |- _ => destruct H
  end.
Ltac inversion_typing :=
  match goal with
  | H : typing _ (var _) _ |- _ => depelim H
  | H : typing _ (abs _ _) _ |- _ => depelim H
  | H : typing _ (app _ _) _ |- _ => depelim H
  | H : typing _ (tabs _ _) _ |- _ => depelim H
  | H : typing _ (tapp _ _) _ |- _ => depelim H
  end.
Ltac inversion_normal :=
  match goal with
  | H : normal (abs _ _) |- _ => depelim H
  | H : normal (tabs _ _) |- _ => depelim H
  | H : normal (app _ _) |- _ => depelim H
  | H : normal (tapp _ _) |- _ => depelim H
  | H : normal (var _) |- _ => clear H
  | H : neutral (abs _ _) |- _ => depelim H
  | H : neutral (tabs _ _) |- _ => depelim H
  | H : neutral (app _ _) |- _ => depelim H
  | H : neutral (tapp _ _) |- _ => depelim H
  | H : neutral (var _) |- _ => clear H
  end.
Ltac inversion_is_abs :=
  match goal with
  | H : is_abs (app _ _) |- _ => depelim H
  | H : is_abs (tapp _ _) |- _ => depelim H
  | H : is_abs (var _) |- _ => depelim H
  | H : ~ is_abs (abs _ _) |- _ => elim H; simp is_abs; trivial
  | H : ~ is_abs (tabs _ _) |- _ => elim H; simp is_abs; trivial
  end.

Ltac learn_all :=
  repeat first [destruct_canonical | inversion_typing | inversion_normal |
    inversion_is_abs].

Ltac my_intuition := intros;
  simpl in *; simp get_var subst remove_var in *; auto;
  learn_all; simpl in *; eauto.

Ltac destruct_ps := repeat
  match goal with
    [ p : env * typ |- _ ] => destruct p; cbn [fst snd] in *; subst
  end.

Ltac subst_lets := repeat
  match goal with
    [ H := _ : _ |- _ ] => subst H
  end.

Require Import Program.Tactics.

Ltac my_obl_tac :=
  program_simpl; subst_lets; unfold pre, post in *; my_intuition.

Ltac specialize_post :=
  match goal with
  | P : forall _ : env * typ, _ /\ canonical _ ?t _ /\ _ -> _,
    H : typing ?e ?t ?T |- _ =>
      specialize (P (e, T)); forward P;
      [ clear P; repeat split; my_intuition
      | destruct P as [[? [? ?]] ?]]
  end; my_intuition.

Ltac postcondition := repeat specialize_post; destruct_ps; repeat split; my_intuition.

Ltac precondition :=
  match goal with
  | H : typing ?e ?t ?T |- exists _ : env * typ, _ /\ canonical _ ?t _ /\ _ =>
      exists (e, T)
  end; repeat split; my_intuition.

Ltac solve_her_order_2 :=
  constructor 2; [intuition | red; simp term_size; auto; omega].

Ltac well_founded :=
  match goal with
  | H : forall _ : typ, _ |- _ =>
      intros; apply H; solve_her_order_2
  end;
  try rewrite kinds_of_tshift; try rewrite depth_tshift; trivial.

Ltac solve_order :=
  try match goal with
  | |- ordtyp _ _ => constructor
  end;
  match goal with
  | H : _ -> _ -> ordtyp (arrow ?T _) ?U |- order ?T ?U =>
    forward H; [|forward H]
  | H : _ -> _ -> ordtyp (arrow _ ?T) ?U |- order ?T ?U =>
    forward H; [|forward H]
  | H : _ -> _ -> ordtyp (all _ ?T) ?U |- order (tsubst ?T _ _) ?U =>
    forward H; [|forward H]
  end; simp is_abs; eauto.

Obligation Tactic := my_obl_tac.


Hint Resolve typing_weakening_kind.
Hint Resolve subst_typ_preserves_typing kinding_strengthening_var.
Hint Resolve neutral_noabs.
Hint Resolve order_ordtyp_trans order_arrow_l order_arrow_r.

Arguments existT [A P] x _.

(* Preconditions *)
Obligation 4. precondition.
  apply typing_weakening_var; auto;
  apply wf_typ_remove_var; apply typing_wf_env in H0; apply H0.
Defined.

Obligation 7. precondition. Defined.
Obligation 9. precondition. Defined.
Obligation 12. repeat specialize_post; precondition. Defined.
Obligation 18. precondition.
  rewrite H; trivial.
Defined.
Obligation 21. precondition. Defined.

(* Well-foundedness *)
Ltac well_founded ::=
  match goal with
  | H : forall _ : typ, _ |- _ =>
    solve_her_order_2
  end;
  try rewrite kinds_of_tshift; try rewrite depth_tshift; trivial.

Obligation Tactic := program_simpl.
Obligation 5. well_founded. Defined.
Obligation 8. well_founded. Defined.
Obligation 10. well_founded. Defined.
Obligation 19. well_founded. Defined.
Obligation 22. well_founded. Defined.
(* Interesting case *)
Obligation 13.
Proof.
  my_obl_tac. postcondition.
  repeat specialize_post. constructor.
  solve_order.
Defined.

(* Post-conditions *)
Obligation 11. my_obl_tac. postcondition. Defined.
Obligation 15. my_obl_tac. postcondition. Defined.
Obligation 16. my_obl_tac. postcondition. Defined.
Obligation 17. my_obl_tac. postcondition.
  econstructor; eauto. rewrite H3; eauto.
Defined.
Obligation Tactic := idtac.
Obligation 20.
  intros; on_call hsubst ltac:(fun c => destruct c).
  my_obl_tac; postcondition.
  rewrite H; trivial.
Defined.
Obligation Tactic := my_obl_tac.
Obligation 23. postcondition. Defined.
Obligation 24. postcondition. Defined.
Obligation 25. postcondition. Defined.
Obligation 27. postcondition.
  econstructor; eauto. rewrite H3; eauto.
Defined.

(* Case (var i), i < X *)
Next Obligation. postcondition.
  - simp subst; compare i X. constructor 2.
  - replace (var i) with (subst (var i) X u) by
    (simp subst; compare i X; trivial).
    eapply subst_preserves_typing; eauto.
Defined.

(* Case (var X) *)
Next Obligation. postcondition.
  - simp subst; compare X X. constructor 2.
  - rewrite H2 in H4; depelim H4. trivial.
  - rewrite H2 in H4; depelim H4. constructor 2.
Defined.

(* Case (var i), i > X *)
Next Obligation.
  destruct i as [|i]; [omega|].
  repeat split.
  - compare (S i) X. simpl. rewrite Nat.sub_0_r. constructor 2.
  - simpl. replace (var i) with (subst (var (S i)) X u) by
    (simp subst; compare (S i) X; simpl; rewrite Nat.sub_0_r; trivial).
    eapply subst_preserves_typing; eauto.
  - my_intuition.
  - my_intuition.
Defined.

Obligation Tactic := idtac.
(* Case (abs T t) *)
Next Obligation.
  intros; on_call hsubst ltac:(fun c => destruct c).
  my_obl_tac; postcondition.
  apply typing_weakening_var; auto.
  apply wf_typ_remove_var. apply typing_wf_env in H1.
  apply H1.
Defined.

(* Case (app t1 t2) *)
Next Obligation.
  intros; on_call hsubst ltac:(fun c => destruct c).
  my_obl_tac; postcondition.
  - abstract (eapply reds_trans; [|eassumption];
    eapply reds_trans; [eapply reds_app; eassumption | eauto 7]).
  - abstract (constructor; solve_order).
Defined.
Obligation Tactic := my_obl_tac.

(* Case (tapp t T) *)
Next Obligation. postcondition.
  - depelim H11; eauto 8. rewrite <- H2; eauto 8.
  - constructor. destruct (regularity H3).
    solve_order. refine (order_ordtyp_trans _ H14).
    constructor. eapply kinds_of_tsubst_all; eauto.
Defined.

(* end hide *)

(** With [hsubst] defined, it is now easy to implement a
    [normalize] function which takes a term and returns its normal form.
    As for [hsubst], we add a precondition and a postcondition
    which allow to show correctness by construction. *)

Definition pre' (t : term) (p : env * typ) : Prop :=
  typing (fst p) t (snd p).
Definition post' (t : term) (n : term) (p : env * typ) : Prop :=
  interp (fst p) t (snd p) n.

Derive Subterm for term.

Equations normalize (t : term) (P : exists (p : env * typ), pre' t p) :
  {n : term | forall (p : env * typ), pre' t p -> post' t n p} by wf t term_subterm :=
normalize (var i) P => var i;
normalize (abs T1 t) P => abs T1 (normalize t _) ;
normalize (app t1 t2) P with normalize t1 _, normalize t2 _ => {
  | exist (abs T t) P1' | exist t2' P2'  => hsubst T t t2' 0 _ ;
  | exist t1' P1' | exist t2' P2' => app t1' t2' };
normalize (tabs k t) P => tabs k (normalize t _) ;
normalize (tapp t T) P with normalize t _ => {
  | exist (tabs k t') P' => subst_typ t' 0 T ;
  | exist t' P' => tapp t' T }.

(* begin hide *)

Ltac my_obl_tac ::=
  program_simpl; unfold pre, post, pre', post' in *; subst_lets; destruct_ps; my_intuition.

Ltac precondition' :=
  match goal with
  | H : typing ?e ?t ?T |- exists _ : env * typ, typing _ ?t _ =>
      exists (e, T)
  end; repeat split; my_intuition.
  
Ltac specialize_post' :=
  match goal with
  | P : forall _ : env * typ, typing _ ?t _ -> _,
    H : typing ?e ?t ?T |- _ =>
      specialize (P (e, T)); forward P;
      [ clear P; repeat split; my_intuition
      | destruct P as [? [? ?]]]
  end; my_intuition.
Ltac postcondition' := repeat specialize_post'; destruct_ps; repeat split; my_intuition.

Solve Obligations with my_obl_tac; precondition'.
Solve Obligations with my_obl_tac; postcondition'.

(* Call to hsubst *)
(* Pre-condition. *)
Next Obligation. my_obl_tac.
  destruct normalize. simpl. red. specialize_post'.
Defined.

Obligation Tactic := idtac.

Next Obligation. my_obl_tac.
  repeat specialize_post'. precondition.
Defined.
(* Post-condition. *)
Next Obligation.
  intros; on_call hsubst ltac:(fun c => destruct c).
  my_obl_tac; repeat specialize_post'; repeat specialize_post.
  repeat split; auto.
  depelim H4; depelim H0; depelim H7; eauto 9.
Defined.

(* Case (app t1 t2) --> (app (tapp t T) t2) *)
Next Obligation. my_obl_tac. postcondition'.
  econstructor. eauto. rewrite H0. econstructor; eauto. auto.
Defined.

(* Case (tapp t T) --> (tapp (all k t) T) *)
Next Obligation. my_obl_tac. destruct normalize. simpl. specialize_post'. postcondition'. Defined.
Next Obligation. my_obl_tac. specialize_post'. postcondition'. depelim H3; eauto 8. Defined.
(* Case (tapp t T) --> (tapp (tapp t T1) T2) *)
Next Obligation. my_obl_tac. postcondition'.
  econstructor; eauto. rewrite H2; eauto.
Defined.

Set Implicit Arguments.
(* end hide *)

(** The existence of the [normalize] function is in itself a proof of the strong
    normalization of Leivant's Predicative System F. *)

Theorem normalization e t T : typing e t T -> exists n, reds t n /\ typing e n T /\ normal n.
Proof.
  intros.
  destruct (normalize t (ex_intro _ (e, T) H)) as [nf P].
  exists nf. exact (P (e, T) H).
Qed.

(* begin hide *)
Print Assumptions normalization.
(* end hide *)

(** ** Consistency

   It is easy to show consistency based on the normalization function.
   We just need lemmas showing that neutral terms cannot inhabit any
   type in an environment with only a type variable, by inversion on the
   neutrality derivation.  *)

Lemma neutral_tvar t k T : neutral t -> typing (etvar empty_env k) t T -> False.
Proof.
  revert T.
  induction t; intros; my_intuition; simpl in *; discriminate.
Qed.

(* begin hide *)
Lemma neutral_empty t T : neutral t -> typing empty_env t T -> False.
Proof.
  revert T.
  induction t; intros; my_intuition; simpl in *; discriminate.
Qed.
(* end hide *)

(** Consistency is then proved using case analysis on an assumed typing
 derivation of falsehood at any universe level [k].  Informally, it is
 showing that [∀ X :* k, X] is not inhabited for any [k]. *)

Corollary consistency k : ~ exists t, typing empty_env t (all k (tvar 0)).
Proof.
  intro. destruct H.
  apply normalization in H.
  destruct H as [nf [_ [Htyp Hnorm]]].
  depelim Hnorm.
  - depelim Htyp.
  - depelim Htyp; depelim Hnorm.
    + depelim Htyp.
    + depelim Htyp.
    + apply (neutral_tvar H Htyp).
  - apply (neutral_empty H Htyp).
Qed.

(* begin hide *)
Print Assumptions consistency.

Extraction "normalize.ml" normalize.
(* end hide *)
