(** printing <= $\Leftarrow$ *)
(** printing <mul $<_{m}$ *)
(** printing <=mul $\le_{m}$ *)
(** printing =mul= $=_{m}$ *)
(** printing {{ $\{$ *)
(** printing }} $\}$ *)
(* begin hide *)
Set Asymmetric Patterns.
Set Implicit Arguments.
Unset Strict Implicit.

Require Import Arith Omega.
Require Import Equations.Equations.

Require Import Fpred.Definitions.

Require Import Relation_Operators Inclusion.

Arguments clos_refl [A].
Hint Constructors clos_refl.
(* end hide *)

(** ** Reduction rules *)

(** To define normalization we must formalize the reduction relation of the calculus.
    Beta-redexes for this calculus are the application of an abstraction
    to a term and the application of a type abstraction to a type. *)

Inductive red : term -> term -> Prop :=
  | E_AppAbs (T : typ) (t1 t2 : term) : red (app (abs T t1) t2) (subst t1 0 t2)
  | E_TappTabs k (T : typ) (t : term) : red (tapp (tabs k t) T) (subst_typ t 0 T).
(* begin hide *)
Hint Constructors red.
(* end hide *)

(** We define the transitive closure of reduction on terms by closing
  [red] by context. *)

Inductive sred : term -> term -> Prop :=
  | Red_sred t t' : red t t' -> sred t t'
  | sred_trans t1 t2 t3 : sred t1 t2 -> sred t2 t3 -> sred t1 t3
  | Par_app_left t1 t1' t2 : sred t1 t1' -> sred (app t1 t2) (app t1' t2)
  | Par_app_right t1 t2 t2' : sred t2 t2' -> sred (app t1 t2) (app t1 t2')
  | Par_abs T t t'  : sred t t' -> sred (abs T t) (abs T t')
  | Par_tapp t t' T : sred t t' -> sred (tapp t T) (tapp t' T)
  | Par_tabs k t t' : sred t t' -> sred (tabs k t) (tabs k t').

(* begin hide *)
Hint Constructors sred.
(* end hide *)

Definition reds t n := clos_refl sred t n.
(* begin hide *)
Hint Unfold reds.

Hint Constructors clos_refl_trans.
(* end hide *)

(** We can prove usual congruence lemmas on [reds] showing that it indeed
  formalizes parallel reduction. *)
(* begin hide *)
Lemma reds_app t u t' u' : reds t t' -> reds u u' -> reds (app t u) (app t' u').
Proof.
  intros. destruct H, H0; red; eauto. 
Qed.

Lemma reds_tapp t U t' : reds t t' -> reds (tapp t U) (tapp t' U).
Proof.
  intros. destruct H; red; eauto. 
Qed.

Lemma reds_trans t u v : reds t u -> reds u v -> reds t v.
Proof.
  intros. destruct H, H0; red; eauto.
Qed.

Lemma reds_tabs T t t' : reds t t' -> reds (tabs T t) (tabs T t').
Proof.
  destruct 1; simpl; eauto.
Qed.

Lemma reds_abs T t t' : reds t t' -> reds (abs T t) (abs T t').
Proof.
  destruct 1; simpl; eauto.
Qed.

Hint Resolve reds_app reds_tapp reds_tabs reds_abs.
(* end hide *)
