Require Import Equations.Equations.
Require Import Arith Omega.

Ltac forward H := 
  match type of H with
    | ?X -> _ => let H' := fresh in assert (H':X); [|specialize (H H'); clear H']
  end.

Ltac depelim_ind C :=
  multimatch goal with
    | H : C |- _ => depelim H
    | H : C _ |- _ => depelim H
    | H : C _ _ |- _ => depelim H
    | H : C _ _ _ |- _ => depelim H
    | H : C _ _ _ _ |- _ => depelim H
    | H : C _ _ _ _ _ |- _ => depelim H
  end.

Definition hidden {A} (a : A) := a.
Ltac hide H := 
  let ty := type of H in
  change ty with (@hidden _ ty) in H.

Lemma UIP_refl_nat (n:nat) (x : n = n) : x = refl_equal.
Proof.
  induction n.
  - change (match 0 as n return 0=n -> Prop with
            | 0 => fun x => x = refl_equal
            | _ => fun _ => True
            end x).
    destruct x; reflexivity.
  - specialize IHn with (f_equal pred x).
    change refl_equal with (f_equal S (@refl_equal _ n)).
    rewrite <- IHn; clear IHn.
    change (match S n as n' return S n = n' -> Prop with
            | 0 => fun _ => True
            | S n' => fun x =>
                x = f_equal S (f_equal pred x)
            end x).
  pattern (S n) at 2 3, x.
  destruct x; reflexivity.
Defined.

Lemma lt_eq_dec_refl x y (p : x = y) : lt_eq_lt_dec x y = inleft (y < x) (right (x < y) p).
Proof. 
  destruct lt_eq_lt_dec. destruct s. elimtype False. subst y. apply (lt_irrefl _ l).
  repeat f_equal. destruct p. apply UIP_refl_nat. (* irrel eq on nat *)
  elimtype False. subst x. apply (lt_irrefl _ l).
Qed.

Ltac compare a b :=
  destruct (lt_eq_lt_dec a b) as [[? | ?] | ?]; try omega.
