
let __ = let rec f _ = Obj.repr f in Obj.repr f

type nat =
| O
| S of nat

type 'a sig0 = 'a
  (* singleton inductive, whose constructor was exist *)

type sumbool =
| Left
| Right

type 'a sumor =
| Inleft of 'a
| Inright

(** val pred : nat -> nat **)

let pred n = match n with
| O -> n
| S u -> u

(** val add : nat -> nat -> nat **)

let rec add n m =
  match n with
  | O -> m
  | S p -> S (add p m)

(** val sub : nat -> nat -> nat **)

let rec sub n m =
  match n with
  | O -> n
  | S k -> (match m with
            | O -> n
            | S l -> sub k l)

(** val lt_eq_lt_dec : nat -> nat -> sumbool sumor **)

let rec lt_eq_lt_dec n m =
  match n with
  | O -> (match m with
          | O -> Inleft Right
          | S _ -> Inleft Left)
  | S n0 -> (match m with
             | O -> Inright
             | S m0 -> lt_eq_lt_dec n0 m0)

(** val le_lt_dec : nat -> nat -> sumbool **)

let rec le_lt_dec n m =
  match n with
  | O -> Left
  | S n0 -> (match m with
             | O -> Right
             | S m0 -> le_lt_dec n0 m0)

(** val le_gt_dec : nat -> nat -> sumbool **)

let le_gt_dec =
  le_lt_dec

(** val pr1 : ('a1 * 'a2) -> 'a1 **)

let pr1 = function
| pr3,_ -> pr3

(** val pr2 : ('a1 * 'a2) -> 'a2 **)

let pr2 = function
| _,pr3 -> pr3

type kind = nat

type typ =
| Tvar of nat
| Arrow of typ * typ
| All of kind * typ

type term =
| Var of nat
| Abs of typ * term
| App of term * term
| Tabs of kind * term
| Tapp of term * typ

(** val tshift : nat -> typ -> typ **)

let rec tshift x = function
| Tvar n -> Tvar (match le_gt_dec x n with
                  | Left -> add (S O) n
                  | Right -> n)
| Arrow (t0, t1) -> Arrow ((tshift x t0), (tshift x t1))
| All (k, t0) -> All (k, (tshift (S x) t0))

(** val shift : nat -> term -> term **)

let rec shift x = function
| Var n -> Var (match le_gt_dec x n with
                | Left -> add (S O) n
                | Right -> n)
| Abs (t0, t1) -> Abs (t0, (shift (S x) t1))
| App (t0, t1) -> App ((shift x t0), (shift x t1))
| Tabs (k, t0) -> Tabs (k, (shift x t0))
| Tapp (t0, t1) -> Tapp ((shift x t0), t1)

(** val shift_typ : nat -> term -> term **)

let rec shift_typ x = function
| Var n -> Var n
| Abs (t0, t1) -> Abs ((tshift x t0), (shift_typ x t1))
| App (t0, t1) -> App ((shift_typ x t0), (shift_typ x t1))
| Tabs (k, t0) -> Tabs (k, (shift_typ (add (S O) x) t0))
| Tapp (t0, t1) -> Tapp ((shift_typ x t0), (tshift x t1))

(** val tsubst : typ -> nat -> typ -> typ **)

let rec tsubst t x t' =
  match t with
  | Tvar n ->
    (match lt_eq_lt_dec n x with
     | Inleft s -> (match s with
                    | Left -> Tvar n
                    | Right -> t')
     | Inright -> Tvar (sub n (S O)))
  | Arrow (t0, t1) -> Arrow ((tsubst t0 x t'), (tsubst t1 x t'))
  | All (k, t0) -> All (k, (tsubst t0 (add (S O) x) (tshift O t')))

(** val subst_typ : term -> nat -> typ -> term **)

let rec subst_typ t x t0 =
  match t with
  | Var n -> Var n
  | Abs (t1, t2) -> Abs ((tsubst t1 x t0), (subst_typ t2 x t0))
  | App (t1, t2) -> App ((subst_typ t1 x t0), (subst_typ t2 x t0))
  | Tabs (k, t1) -> Tabs (k, (subst_typ t1 (add (S O) x) (tshift O t0)))
  | Tapp (t1, t2) -> Tapp ((subst_typ t1 x t0), (tsubst t2 x t0))

(** val hsubst : typ -> term -> term -> nat -> term **)

let hsubst a a0 a1 a2 =
  let rec fix_F x =
    let u = pr1 x in
    let u0 = pr1 (pr2 (pr2 x)) in
    let x0 = pr1 (pr2 (pr2 (pr2 x))) in
    let hsubst0 = fun a3 a4 a5 a6 ->
      let y = a3,(a4,(a5,(a6,__))) in (fun _ -> fix_F y)
    in
    (match pr1 (pr2 x) with
     | Var n ->
       (match lt_eq_lt_dec n x0 with
        | Inleft s -> (match s with
                       | Left -> Var n
                       | Right -> u0)
        | Inright -> Var (pred n))
     | Abs (t, t0) -> Abs (t, (hsubst0 u t0 (shift O u0) (S x0) __))
     | App (t, t0) ->
       (match hsubst0 u t u0 x0 __ with
        | Abs (t1, t2) -> hsubst0 t1 t2 (hsubst0 u t0 u0 x0 __) O __
        | x1 -> App (x1, (hsubst0 u t0 u0 x0 __)))
     | Tabs (k, t) ->
       Tabs (k, (hsubst0 (tshift O u) t (shift_typ O u0) x0 __))
     | Tapp (t, t0) ->
       (match hsubst0 u t u0 x0 __ with
        | Tabs (_, t1) -> subst_typ t1 O t0
        | x1 -> Tapp (x1, t0)))
  in fix_F (a,(a0,(a1,(a2,__))))

(** val normalize : term -> term **)

let normalize a =
  let rec fix_F x =
    let normalize0 = fun a0 -> let y = a0,__ in (fun _ -> fix_F y) in
    (match pr1 x with
     | Var n -> Var n
     | Abs (t, t0) -> Abs (t, (normalize0 t0 __))
     | App (t, t0) ->
       (match normalize0 t __ with
        | Abs (t1, t2) -> hsubst t1 t2 (normalize0 t0 __) O
        | x0 -> App (x0, (normalize0 t0 __)))
     | Tabs (k, t) -> Tabs (k, (normalize0 t __))
     | Tapp (t, t0) ->
       (match normalize0 t __ with
        | Tabs (_, t1) -> subst_typ t1 O t0
        | x0 -> Tapp (x0, t0)))
  in fix_F (a,__)
