(* begin hide *)
From Equations Require Import Equations.
From EquationsReloaded Require Import Fin.

Inductive vector (A : Type) : nat -> Type :=
| nil : vector A 0
| cons (a : A) n (v : vector A n) : vector A (S n).

Inductive fin : nat -> Set :=
| fz : forall n, fin (S n)
| fs : forall n, fin n -> fin (S n).

Derive Signature for fin.

Equations nth {A n} (v : vector A n) (f : fin n) : A :=
nth (cons x _ _) (fz n) := x;
nth (cons _ ?(n) v) (fs n f) := nth v f.

Arguments nil {A}. Arguments cons {A} a {n} v.
(* end hide *)

Check nth_equation_1 : forall (A : Type) (f : fin 0), ImpossibleCall (nth nil f).
Check nth_equation_2 : forall (A : Type) (n : nat) (a : A) (v : vector A n), nth (cons a v) (fz n) = a.
Check nth_equation_3 : forall A n a v f, nth (cons a v) (fs n f) = nth v f.

Check nth_elim : forall P : forall (A : Type) (n : nat), vector A n -> fin n -> A -> Prop,
    (forall A n a v, P A (S n) (cons a v) (fz n) a) ->
    (forall A n a v f, P A n v f (nth v f) -> P A (S n) (cons a v) (fs n f) (nth v f)) ->
    forall A n v f, P A n v f (nth v f).
